# 
# Copyright (C) 2013  Smithsonian Astrophysical Observatory
# 
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""

Summary
=======

This module allows basic access to the CIAO proposal tools from
Python. At present the only supported tool is colden.

>>> import proptools
>>> proptools.colden(23, 40)
6.56
>>> proptools.colden(23, 40, dataset='bell')
6.66
>>> proptools.colden([2.1, 15.8, 356.2, 178.9], [43.5, -0.1, 83.4, -76])
[9.0, 3.17, 9.65, 10.04]
>>> proptools.colden([2.1, 15.8, 356.2, 178.9], [43.5, -0.1, 83.4, -76], dataset='bell')
[7.34, 3.35, 10.61, None]

The return value from colden has units of 10^20 cm^-2

"""

import subprocess as sbp
import numpy as np

__all__ = ("colden", )

from ciao_contrib.logger_wrapper import initialize_module_logger

logger = initialize_module_logger("proptools")

v2 = logger.verbose2
v3 = logger.verbose3
v4 = logger.verbose4
v5 = logger.verbose5

def _colden(ra, dec, dataset):
    """Returns the column density, in units of 10^20 cm^-2,
    for the given location and dataset. ra and dec are
    scalars, and are decimal degrees (J2000). dataset should
    be NRAO or Bell.

    None is returned if the value can not be computed. This
    includes invalid inputs or there is a problem parsing the
    output.
    """

    v2("Running colden with ra={} dec={} for dataset {}".format(ra, dec, dataset))
    proc = sbp.Popen(["prop_colden", "f", "j/deg", "data",
                      dataset, "eval", str(ra), str(dec)],
                      stdout=sbp.PIPE, stderr=sbp.PIPE)
    (sout, serr) = proc.communicate()
    rval = proc.returncode
    v4("* colden output: returncode={}".format(rval))
    v4("* colden output: stdout={}".format(sout))
    v4("* colden output: stderr={}".format(serr))
    
    if rval != 0:
        v2("Colden failed to run successfully")
        return None

    lines = sout.splitlines()
    try:
        oline = lines[-2]
    except IndexError:
        v2("Unexpected colden output: not enough lines")
        return None

    toks = oline.split()
    out = None
    try:
        if toks[0] == 'Hydrogen' and toks[1] == 'density':
            out = float(toks[4])
        else:
            v2("Unexpected colden output; line should start 'Hydrogen density', not '{} {}'".format(toks[0], toks[1]))

    except IndexError:
        v2("Unexpected colden output: line should start 'Hydrogen density', not '{}'".format(oline))

    except ValueError:
        v2("Unexpected colden output: unable to convert '{}' to a number".format(toks[4]))

    return out

def colden(ra, dec, dataset='NRAO'):
    """Return the column density, in units of 10^20 cm^2,
    for the given location(s).

    ra and dec are in decimal degrees, J2000. The dataset argument
    should be 'NRAO' or 'Bell' (case insensitive).

    If ra and dec are scalar, then a scalar value is returned.
    If they are arrays, of the same length, then an array is returned
    (if either ra or dec is a numpy array then the return will be a
    numpy array - which may have a surprising data type if
    any values are missing - otherwise it is a Python list).

    A value of None is returned for any position for which a value can
    not be computed.

    There is no support for changing the velocity limits for the
    Stark et al data (dataset='Bell') or for indicating whether
    the result is interpolated or not.
    """

    if dataset.upper() not in ["NRAO", "BELL"]:
        raise ValueError("Invalid dataset={}; must be NRAO or Bell".format(dataset))

    try:
        args = zip(ra,dec)
        multi = True

    except TypeError:
        args = zip([ra], [dec])
        multi = False

    if len(args) == 1:
        v2("Running colden for 1 position (using {} data set)".format(dataset))
    else:
        v2("Running colden for {} positions (using {} data set)".format(len(args), dataset))
        
    out = []
    for (a,b) in args:
        out.append(_colden(a, b, dataset))

    if multi:
        if isinstance(ra, np.ndarray) or isinstance(dec, np.ndarray):
            oval = np.asarray(out)
        else:
            oval = out
    else:
        oval = out[0]

    return oval
